/*
This code is a modified version of chull.c, as described in 
"Computational Geometry in C" (Second Edition), Chapter 4,
modified for use with the shortest paths code sp.cpp (thus the
name chullsp).  It is not written to be comprehensible without the 
explanation in that book.

Input: 3n integer coordinates for the points.

It is used for the purposes of the shortest paths code in three
different ways:
	1. -o flag:  to output a geomview file of the polytope.
	2. no flag:  to produce a .in file for sp.
	3  -g flag:  to postprocess the sp output to geomview.
A fourth option is:
	4. -m flag:  mathematica rather than geomview output.
Finally, a fifth option is
	5. -n flag:  to indicate nonconvex (e.g., terrain) data
It is clear that these flag names are not well chosen, and should
be better organized.

The reason all this is done in the convex hull code is that chullsp creates
the data structure, from which it is then easy to tack on different
printout functions.  But this is all rather hacked together in a rather
unsatisfactory manner at the moment.

Compile:	gcc -o chullsp chullsp.c -lm
or	 	make chullsp

Written by Joseph O'Rourke, with contributions by several students.
Modifications for use with sp by Biliana Kaneva.
Last modified: Septemeber 2000
Questions to orourke@cs.smith.edu.
--------------------------------------------------------------------
This code is Copyright 2000 by Joseph O'Rourke.  It may be freely 
redistributed in its entirety provided that this copyright notice is 
not removed.
--------------------------------------------------------------------
*/
#include <stdio.h>
#include <math.h>

/*Define Boolean type */
typedef	enum { FALSE, TRUE }	bool;

/* Define vertex indices. */
#define X   0
#define Y   1
#define Z   2
#define MAX_INT   2147483647 

/* Define structures for vertices, edges and faces */
typedef struct tVertexStructure tsVertex;
typedef tsVertex *tVertex;

typedef struct tEdgeStructure tsEdge;
typedef tsEdge *tEdge;

typedef struct tFaceStructure tsFace;
typedef tsFace *tFace;

struct tVertexStructure {
   int      v[4];
   int	    vnum;
   tEdge    duplicate;	        /* pointer to incident cone edge (or NULL) */
   bool     onhull;		/* T iff point on hull. */
   bool	    mark;		/* T iff point already processed. */
   bool     visible;
   tVertex  next, prev;
};

struct tEdgeStructure {
   tFace    adjface[2];
   tVertex  endpts[2];
   tFace    newface;            /* pointer to incident cone face. */
   bool     delete;		/* T iff edge should be delete. */
   tEdge    next, prev;
};

struct tFaceStructure {
   tEdge    edge[3];
   tVertex  vertex[3];
   int      label;
   bool	    visible;	        /* T iff face visible from new point. */
   bool     lower;
   tFace    next, prev;
};

typedef struct PathStruct {

  int num;
  int *faces;
  double *fractions;
  int *vlabel;
  int last;
} Path;

/* Define flags */
#define ONHULL   	TRUE
#define REMOVED  	TRUE
#define VISIBLE  	TRUE
#define PROCESSED	TRUE
#define SAFE		1000000		/* Range of safe coord values. */

/* Global variable definitions */
Path *path;
tVertex vertices = NULL;
tEdge edges    	 = NULL;
tFace faces    	 = NULL;
double s[3];
bool debug = FALSE;
bool check = FALSE;
bool output = FALSE;
bool geomview = FALSE;
bool nonconvex = FALSE;
bool mathematica = FALSE;
int  vnum = 0;
int vstart[3];
long vindex;

/* Function declarations */
tVertex MakeNullVertex( void );
void    ReadVertices( void );
void    Print( void );
void    SubVec( int a[3], int b[3], int c[3]);
void    DoubleTriangle( void );
void    ConstructHull( void );
bool	AddOne( tVertex p );
int     VolumeSign(tFace f, tVertex p);
int 	Volumei( tFace f, tVertex p );
tFace	MakeConeFace( tEdge e, tVertex p );
void    MakeCcw( tFace f, tEdge e, tVertex p );
tEdge   MakeNullEdge( void );
tFace   MakeNullFace( void );
tFace   MakeFace( tVertex v0, tVertex v1, tVertex v2, tFace f );
void    CleanUp( void );
void    CleanEdges( void );
void    CleanFaces( void );
void    CleanVertices( void );
bool	Collinear( tVertex a, tVertex b, tVertex c );
void    CheckEuler(int V, int E, int F );
int     Normz(tFace f);
void	PrintPoint( tVertex p );
void    Checks( void );
void	Consistency( void );
void	Convexity( void );
void	PrintOut( tVertex v );
void	PrintVertices( void );
void	PrintEdges( void );
void	PrintFaces( void );
void    PrintInputFile( void );
void    PrintGeomviewPolytope( void );
void    PrintGeomviewSP( void );
void    PrintMathematica(void);
void    PrintMathematicaFlip(void);
void    LowerFaces();

#include "macros.h"

/*-------------------------------------------------------------------*/
main( int argc, char *argv[] )
{

  if ( argc > 1 && argv[1][0] == '-' ) {
    if( argv[1][1] ==  'd' ) {
      debug = TRUE;
      check = TRUE;
      fprintf( stderr, "Debug and check mode\n");
    }
    if( argv[1][1] == 'c' ) {
      check = TRUE;
      fprintf( stderr, "Check mode\n");
    }
    if (argv[1][1] == 'o') {
      output = TRUE;
    }

    if (argv[1][1] == 'g') {
      geomview = TRUE;
    }
    if (argv[1][1] == 'm') {
      mathematica = TRUE;
    }
    if (argv[1][2] == 'n') {
      nonconvex = TRUE;
    }
  }
  else if ( argc > 1 && argv[1][0] != '-' ) {
    printf ("Usage:  %s -d[ebug] c[heck] o[utput geomview format] g[eomview shortest paths]\n", *argv );
    printf ("x y z coords of vertices from stdin\n");
    exit(1);
  }
  
   ReadVertices();
   DoubleTriangle();
   ConstructHull();

   if (nonconvex)
     {
       LowerFaces();
       
     }

   if (output)
     PrintGeomviewPolytope();
   else
     if (geomview)
       PrintGeomviewSP();
     else if (mathematica)
       PrintMathematica();
   /*       PrintMathematicaFlip();*/
     else
       PrintInputFile();
   /* Print();*/
}


void    LowerFaces( void )
{
   tFace f = faces;
   tVertex  v;
   int i;
   int   Flower = 0;   /* Total number of lower faces. */

   do {
     if ( Normz( f ) < 0 ) {
        Flower++;
        f->lower = TRUE;
     }
     else f->lower = FALSE;
     f = f->next;
   } while ( f != faces );

   v = vertices;

   do {
     v->v[Z] = v->v[3];
     v = v->next;
   } while (v!=vertices);
   
   /*printf("A total of %d lower faces identified.\n", Flower);*/
}

/*---------------------------------------------------------------------
MakeNullVertex: Makes a vertex, nulls out fields.
---------------------------------------------------------------------*/
tVertex	MakeNullVertex( void )
{
   tVertex  v;
   
   NEW( v, tsVertex );
   v->duplicate = NULL;
   v->onhull = !ONHULL;
   v->mark = !PROCESSED;
   ADD( vertices, v );

   return v;
}

/*---------------------------------------------------------------------
ReadVertices: Reads in the vertices, and links them into a circular
list with MakeNullVertex.  There is no need for the # of vertices to be
the first line: the function looks for EOF instead.  Sets the global
variable vertices via the ADD macro.
---------------------------------------------------------------------*/
void	ReadVertices( void )
{
   tVertex  v;
   int      x, y, z, i,j;
   int      vnum_local = 0;

   if (geomview || mathematica)
     {
        scanf("%d", &vnum);

        path = (Path *)calloc(vnum, sizeof(Path));
        for (i=0; i < vnum; i++) {

          scanf("%d", &path[i].num);

          path[i].faces = (int *)calloc((path[i].num+1), sizeof(int));
          path[i].fractions = (double *)calloc((path[i].num+1), sizeof(double));     
          path[i].vlabel = (int *)calloc((path[i].num+1), sizeof(int));
          for (j=0; j < path[i].num; j++)
            {
              scanf("%d %d %lf", &path[i].faces[j], &path[i].vlabel[j], &path[i].fractions[j]);
            }
          path[i].last = i;
        }        
        scanf("%d", &vindex);
     }

   /*   scanf("%d %d %d", &vstart[0], &vstart[1], &vstart[2]);*/

   while ( scanf ("%d %d %d", &x, &y, &z ) != EOF )  {
      v = MakeNullVertex();
      v->v[X] = x;
      v->v[Y] = y;
      if (nonconvex)
        {
          v->v[3] = z;
          v->v[Z] = x*x + y*y;
        }
      else
        v->v[Z] = z;

      v->vnum = vnum_local++;
      if ( ( abs(x) > SAFE ) || ( abs(y) > SAFE ) || ( abs(z) > SAFE ) ) {
         printf("Coordinate of vertex below might be too large: run with -c flag\n");
         PrintPoint(v);
      }
   }



}

/*---------------------------------------------------------------------------
 PrintGeomview - generates the geomview file for displaying the shortest
 paths
  ---------------------------------------------------------------------------*/
void PrintGeomviewSP(void)
{
  int numv=0, i, j,k,m, F=0;
  double source[3], V[3];
  /* Pointers to vertices, edges, faces. */
  tVertex  v;
  tEdge    e;
  tFace    f,f1, f2;
  int VC=0;

  v = vertices;
   do {                                 
      if( v->mark ) v->vnum=VC++;           
      if (v->vnum == vindex)
        {
          for (i = 0; i < 3; i++)
            vstart[i]=v->v[i];
        }
      v = v->next;

   } while ( v != vertices );
  printf("VECT\n");
  
  for (i=0; i < vnum; i++)
    numv+=(path[i].num+1);

  printf("%d %d 1\n", vnum, numv);

  for (i=0; i < vnum; i++)
    printf("%d ", (path[i].num+1));
  printf("\n");

  printf("1 ");
  for (i=1; i < vnum; i++)
    printf("0 ");
  printf("\n");

  f = faces;
  do {
    if (!nonconvex || f->lower)   
      {
        f->label = F;
        ++F;                              
      }
    else
      f->label = -1;

   
    f  = f ->next;
  } while ( f  != faces );

  
  source[X] = vstart[X];
  source[Y] = vstart[Y];
  source[Z] = vstart[Z];
  
  for (i=0; i < vnum; i++)
    {
      printf("%lf %lf %lf\n", source[X],source[Y],source[Z]);
      for (j=1; j < path[i].num; j++)
	{
	  f1=faces;
	  do {
	    f1=f1->next;
	  }while (f1->label != path[i].faces[j-1]);
	  f2=faces;
	  do {
	    f2=f2->next;
	  } while(f2->label != path[i].faces[j]);



	  for (k=0; k < 3; k++)
	    for (m=0; m < 3; m++)
	      if (f1->edge[k] == f2->edge[m])
		{
		  e=f2->edge[m];
		  break;
		}

          if (path[i].fractions[j] != 0)
            {
	  if (e->endpts[0]->vnum == path[i].vlabel[j])
	    {
	      V[X] = e->endpts[0]->v[X] + path[i].fractions[j]*
		(e->endpts[1]->v[X] - e->endpts[0]->v[X]);
	      V[Y] = e->endpts[0]->v[Y] + path[i].fractions[j]*
		(e->endpts[1]->v[Y] - e->endpts[0]->v[Y]);
	      V[Z] = e->endpts[0]->v[Z] + path[i].fractions[j]*
		(e->endpts[1]->v[Z] -  e->endpts[0]->v[Z]);
	    }
	  else
	    {
	      V[X] = e->endpts[1]->v[X] + path[i].fractions[j]*
		(e->endpts[0]->v[X] - e->endpts[1]->v[X]);
	      V[Y] = e->endpts[1]->v[Y] + path[i].fractions[j]*
		(e->endpts[0]->v[Y] - e->endpts[1]->v[Y]);
	      V[Z] = e->endpts[1]->v[Z] + path[i].fractions[j]*
		(e->endpts[0]->v[Z] - e->endpts[1]->v[Z]);
	    }
            }
	  else
            {
              v = vertices;
              /*              printf("vert num %d\n", path[i].vlabel[j]);*/
              while (v->vnum != path[i].vlabel[j])
                {
                  v = v->next;
                } 
                for (k = 0; k < 3; k++)
                  V[k] = v->v[k];
            }
	  printf("%lf %lf %lf\n", V[X], V[Y], V[Z]);
	}

      v=vertices;
      do {
	v=v->next;
      }while (v->vnum != path[i].last);


      printf("%d %d %d\n", v->v[X], v->v[Y], v->v[Z]);
    }
  printf("0 0 0 1\n");

  

}



void PrintMathematica(void)
{
  int numv=0, i, j,k,m, F=0;
  double source[3], V[3];
  /* Pointers to vertices, edges, faces. */
  tVertex  v;
  tEdge    e;
  tFace    f,f1, f2;
  int VC=0;
  fprintf(stderr, "In Mathematica\n");
  printf("Off[General::spell];\n");
  
  v = vertices;
   do {                                 
      if( v->mark )
        {
          v->vnum=VC++;           
          if (v->vnum == vindex)
            {
              for (i = 0; i < 3; i++)
                vstart[i]=v->v[i];
            }
          printf("p%d = {%d, %d, %d};\n", v->vnum, v->v[X], v->v[Y],v->v[Z]);
        }
      
      v = v->next;

   } while ( v != vertices );

  
  for (i=0; i < vnum; i++)
    numv+=(path[i].num+1);

  printf("f[x_, y_, z_] := Polygon[{x,y,z}]\n");
  printf("Show[Graphics3D[{EdgeForm[],Thickness[0.0001],");
  f = faces;
  do {
    if (!nonconvex || f->lower)   
      {
        f->label = F;
        ++F;               
        printf("f[p%d, p%d, p%d],\n", f->vertex[0]->vnum,f->vertex[1]->vnum,f->vertex[2]->vnum);
      }
    else
      f->label = -1;

   
    f  = f ->next;
  } while ( f  != faces );
 
  
  source[X] = vstart[X];
  source[Y] = vstart[Y];
  source[Z] = vstart[Z];
  
  for (i=0; i < vnum; i++)
    {
      printf("Line[{{%lf, %lf, %lf},", source[X],source[Y],source[Z]);
      for (j=1; j < path[i].num; j++)
	{
	  f1=faces;
	  do {
	    f1=f1->next;
	  }while (f1->label != path[i].faces[j-1]);
	  f2=faces;
	  do {
	    f2=f2->next;
	  } while(f2->label != path[i].faces[j]);



	  for (k=0; k < 3; k++)
	    for (m=0; m < 3; m++)
	      if (f1->edge[k] == f2->edge[m])
		{
		  e=f2->edge[m];
		  break;
		}

          if (path[i].fractions[j] != 0)
            {
	  if (e->endpts[0]->vnum == path[i].vlabel[j])
	    {
	      V[X] = e->endpts[0]->v[X] + path[i].fractions[j]*
		(e->endpts[1]->v[X] - e->endpts[0]->v[X]);
	      V[Y] = e->endpts[0]->v[Y] + path[i].fractions[j]*
		(e->endpts[1]->v[Y] - e->endpts[0]->v[Y]);
	      V[Z] = e->endpts[0]->v[Z] + path[i].fractions[j]*
		(e->endpts[1]->v[Z] -  e->endpts[0]->v[Z]);
	    }
	  else
	    {
	      V[X] = e->endpts[1]->v[X] + path[i].fractions[j]*
		(e->endpts[0]->v[X] - e->endpts[1]->v[X]);
	      V[Y] = e->endpts[1]->v[Y] + path[i].fractions[j]*
		(e->endpts[0]->v[Y] - e->endpts[1]->v[Y]);
	      V[Z] = e->endpts[1]->v[Z] + path[i].fractions[j]*
		(e->endpts[0]->v[Z] - e->endpts[1]->v[Z]);
	    }
            }
	  else
            {
              v = vertices;
              /*              printf("vert num %d\n", path[i].vlabel[j]);*/
              while (v->vnum != path[i].vlabel[j])
                {
                  v = v->next;
                } 
                for (k = 0; k < 3; k++)
                  V[k] = v->v[k];
            }
	  printf("{%lf, %lf, %lf},\n", V[X], V[Y], V[Z]);
	}

      v=vertices;
      do {
	v=v->next;
      }while (v->vnum != path[i].last);


      printf("{%d, %d, %d}}]", v->v[X], v->v[Y], v->v[Z]);
      if (i != vnum-1)
        printf(",\n");
    }
  printf("}, {PlotRange->All, Boxed->False, ");
  printf("LightSources->{{{1.3, -2.4, 2}, GrayLevel[0.3]},");
  printf("{{2.7,0,2}, GrayLevel[0.3]}, {{-2.3,-1.4, 2},GrayLevel[0.4]}}}]]");
}


void PrintMathematicaFlip(void)
{
  int numv=0, i, j,k,m, F=0;
  double source[3], V[3];
  /* Pointers to vertices, edges, faces. */
  tVertex  v;
  tEdge    e;
  tFace    f,f1, f2;
  int VC=0;
  fprintf(stderr, "In Mathematica\n");
  printf("Off[General::spell];\n");
  
  v = vertices;
   do {                                 
      if( v->mark )
        {
          v->vnum=VC++;           
          if (v->vnum == vindex)
            {
              for (i = 0; i < 3; i++)
                vstart[i]=v->v[i];
            }
          printf("p%d = {%d, %d, %d};\n", v->vnum, v->v[X], v->v[Y],(0-v->v[Z]));
        }
      
      v = v->next;

   } while ( v != vertices );

  
  for (i=0; i < vnum; i++)
    numv+=(path[i].num+1);

  printf("f[x_, y_, z_] := Polygon[{x,y,z}]\n");
  printf("Show[Graphics3D[{EdgeForm[],Thickness[0.0001],");
  f = faces;
  do {
    if (!nonconvex || f->lower)   
      {
        f->label = F;
        ++F;               
        printf("f[p%d, p%d, p%d],\n", f->vertex[0]->vnum,f->vertex[1]->vnum,f->vertex[2]->vnum);
      }
    else
      f->label = -1;

   
    f  = f ->next;
  } while ( f  != faces );
 
  
  source[X] = vstart[X];
  source[Y] = vstart[Y];
  source[Z] = vstart[Z];
  
  for (i=0; i < vnum; i++)
    {
      printf("Line[{{%lf, %lf, %lf},", source[X],source[Y],(0-source[Z]));
      for (j=1; j < path[i].num; j++)
	{
	  f1=faces;
	  do {
	    f1=f1->next;
	  }while (f1->label != path[i].faces[j-1]);
	  f2=faces;
	  do {
	    f2=f2->next;
	  } while(f2->label != path[i].faces[j]);



	  for (k=0; k < 3; k++)
	    for (m=0; m < 3; m++)
	      if (f1->edge[k] == f2->edge[m])
		{
		  e=f2->edge[m];
		  break;
		}

          if (path[i].fractions[j] != 0)
            {
	  if (e->endpts[0]->vnum == path[i].vlabel[j])
	    {
	      V[X] = e->endpts[0]->v[X] + path[i].fractions[j]*
		(e->endpts[1]->v[X] - e->endpts[0]->v[X]);
	      V[Y] = e->endpts[0]->v[Y] + path[i].fractions[j]*
		(e->endpts[1]->v[Y] - e->endpts[0]->v[Y]);
	      V[Z] = e->endpts[0]->v[Z] + path[i].fractions[j]*
		(e->endpts[1]->v[Z] -  e->endpts[0]->v[Z]);
	    }
	  else
	    {
	      V[X] = e->endpts[1]->v[X] + path[i].fractions[j]*
		(e->endpts[0]->v[X] - e->endpts[1]->v[X]);
	      V[Y] = e->endpts[1]->v[Y] + path[i].fractions[j]*
		(e->endpts[0]->v[Y] - e->endpts[1]->v[Y]);
	      V[Z] = e->endpts[1]->v[Z] + path[i].fractions[j]*
		(e->endpts[0]->v[Z] - e->endpts[1]->v[Z]);
	    }
            }
	  else
            {
              v = vertices;
              /*              printf("vert num %d\n", path[i].vlabel[j]);*/
              while (v->vnum != path[i].vlabel[j])
                {
                  v = v->next;
                } 
                for (k = 0; k < 3; k++)
                  V[k] = v->v[k];
            }
	  printf("{%lf, %lf, %lf},\n", V[X], V[Y], (0-V[Z]));
	}

      v=vertices;
      do {
	v=v->next;
      }while (v->vnum != path[i].last);


      printf("{%d, %d, %d}}]", v->v[X], v->v[Y], (0-v->v[Z]));
      if (i != vnum-1)
        printf(",\n");
    }
  printf("}, {PlotRange->All, Boxed->False, ");
  printf("LightSources->{{{1.3, -2.4, 2}, GrayLevel[0.3]},");
  printf("{{2.7,0,2}, GrayLevel[0.3]}, {{-2.3,-1.4, 2},GrayLevel[0.4]}}}]]");
}



/*-------------------------------------------------------------------
 PrintInputFiles - generates the input file for SP
 --------------------------------------------------------------------*/
void    PrintInputFile( void )
{

   /* Pointers to vertices, edges, faces. */
   tVertex  v;
   tEdge    e;
   tFace    f;
   double x,y,z,d1, d2, V[3][2];
   int xmin, ymin, xmax, ymax;
   int i,j,k;
   /* Counters for Euler's formula. */
   int 	VC = 0, E = 0 , F = 0;
   int  sface = -1, svertex = -1;

   v = vertices;
   do {                                 
      if( v->mark ) v->vnum=VC++;           
      
      v = v->next;
   } while ( v != vertices );
   
   f = faces;
   do {

     if (!nonconvex || f->lower)
       {
         f->label = F;
         ++F;         
       }
     else
       f->label = -1;
     f  = f ->next;
   } while ( f  != faces );
   
   printf("%d\n%d\n", F,VC);



   f = faces;
   do {

     if (!nonconvex || f->lower)
       {
         /*
         if (sface == -1)
           {
             for (i=0; i < 3; i++)
               {
                 if (f->vertex[i]->v[X] == vstart[X] && 
                     f->vertex[i]->v[Y] == vstart[Y] &&
                     f->vertex[i]->v[Z] == vstart[Z])
                   {
                     sface = f->label;
                     svertex = i;
                     break;
                   }
                     
               }
           }
         */
         /* take this out*/
         V[0][X]=V[0][Y]=V[1][Y]=0;
         V[1][X]=sqrt((f->vertex[1]->v[X]-f->vertex[0]->v[X])*
		  (f->vertex[1]->v[X]-f->vertex[0]->v[X])+
		  (f->vertex[1]->v[Y]-f->vertex[0]->v[Y])*
		  (f->vertex[1]->v[Y]-f->vertex[0]->v[Y])+
		  (f->vertex[1]->v[Z]-f->vertex[0]->v[Z])*
		  (f->vertex[1]->v[Z]-f->vertex[0]->v[Z]));

         d1 =((f->vertex[2]->v[X]-f->vertex[0]->v[X])*
		  (f->vertex[2]->v[X]-f->vertex[0]->v[X])+
		  (f->vertex[2]->v[Y]-f->vertex[0]->v[Y])*
		  (f->vertex[2]->v[Y]-f->vertex[0]->v[Y])+
		  (f->vertex[2]->v[Z]-f->vertex[0]->v[Z])*
		  (f->vertex[2]->v[Z]-f->vertex[0]->v[Z]));

         d2 = ((f->vertex[2]->v[X]-f->vertex[1]->v[X])*
               (f->vertex[2]->v[X]-f->vertex[1]->v[X])+
		  (f->vertex[2]->v[Y]-f->vertex[1]->v[Y])*
               (f->vertex[2]->v[Y]-f->vertex[1]->v[Y])+
               (f->vertex[2]->v[Z]-f->vertex[1]->v[Z])*
               (f->vertex[2]->v[Z]-f->vertex[1]->v[Z]));
         V[2][X] = (V[1][X]*V[1][X] + d1 - d2)/(2*V[1][X]);
         V[2][Y] = sqrt(d1 - V[2][X]*V[2][X]);

         

         printf("%1.17lf %1.17lf %d %1.17lf %1.17lf %d %1.17lf %1.17lf %d\n",
                V[0][X], V[0][Y],
                f->vertex[0]->vnum,
                V[1][X], V[1][Y],
                f->vertex[1]->vnum,
                V[2][X], V[2][Y],
	     f->vertex[2]->vnum );

       }
      f = f->next;
   } while ( f != faces );  

   f = faces;
   do {

     if (!nonconvex || f->lower)
       {
         for (k=0; k < 3; k++) 
           {
             for (i=0; i < 3; i++)
               if ((f->edge[i]->endpts[0]->vnum == f->vertex[k]->vnum &&
                    f->edge[i]->endpts[1]->vnum == f->vertex[(k+1)%3]->vnum) ||
                   (f->edge[i]->endpts[1]->vnum == f->vertex[k]->vnum &&
                    f->edge[i]->endpts[0]->vnum == f->vertex[(k+1)%3]->vnum))
                 break;
             
             if (f->edge[i]->adjface[0]->label == f->label)
               {
                 

                 for (j=0; j < 3; j++)
                   if ((f->edge[i]->endpts[0] == f->edge[i]->adjface[1]->vertex[j] &&
                        f->edge[i]->endpts[1] == f->edge[i]->adjface[1]->vertex[(j+1)%3])||
                       (f->edge[i]->endpts[0] == f->edge[i]->adjface[1]->vertex[(j+1)%3] &&
                        f->edge[i]->endpts[1] == f->edge[i]->adjface[1]->vertex[j]))
                     {
                       if (!nonconvex || f->edge[i]->adjface[1]->lower)
                         {
                           printf("%d ", f->edge[i]->adjface[1]->label);
                           printf("%d\n",j);
                         }
                       else
                         {
                           printf("-1 -1\n");
                         }
                     }
               }
             
             else
               {

                 for (j=0; j < 3; j++)
                   if ((f->edge[i]->endpts[0] == f->edge[i]->adjface[0]->vertex[j] &&
                        f->edge[i]->endpts[1] == f->edge[i]->adjface[0]->vertex[(j+1)%3])||
                       (f->edge[i]->endpts[0] == f->edge[i]->adjface[0]->vertex[(j+1)%3] &&
                        f->edge[i]->endpts[1] == f->edge[i]->adjface[0]->vertex[j]))
                     {
                       if (!nonconvex || f->edge[i]->adjface[0]->lower)
                         {
                           printf("%d ", f->edge[i]->adjface[0]->label);
                           printf("%d\n",j);
                         }
                       else
                         printf("-1 -1\n");
                     }

               }
	   }
       }
     
   

     f = f->next;
   } while ( f != faces );  


   //   printf("%d\n%d\n", sface, svertex);



}


/*-----------------------------------------------------------------------------------
  PrintGeomviewPolytope - creates the input file for geomview to display the
  polytope whose shortest path we are trying to find
*/
void    PrintGeomviewPolytope( void )
{

   /* Pointers to vertices, edges, faces. */
   tVertex  v;
   tEdge    e;
   tFace    f;
   double x,y,z,d1, d2, V[3][2];
   int xmin, ymin, xmax, ymax;
   int i,j;
   /* Counters for Euler's formula. */
   int 	VC = 0, E = 0 , F = 0;

   printf("OFF\n");

   v = vertices;
   do {         
      if( v->mark ) v->vnum=VC++;           
      v = v->next;
   } while ( v != vertices );
   
   f = faces;
   do {
     f->label = F;
     if (!nonconvex || f->lower)
       ++F;
     f  = f ->next;
   } while ( f  != faces );
   
	
   /* Edges. */	
   e = edges;
   do {
     if (nonconvex)
       {
         if (e->adjface[0]->lower || e->adjface[1]->lower)
           E++;
       }
     else
       E++;
      e = e->next;
   } while ( e != edges );
   printf("%d %d %d\n",VC, F, E);

   v = vertices;
   do {
     v=v->next;
   } while (v->vnum != 0);

   do {    
     printf( "%d %d %d\n",
             v->v[X], v->v[Y], v->v[Z] );
     v = v->next;
   } while ( v->vnum != 0 );

   f = faces;

   do {

     if (!nonconvex || f->lower)
           printf("3 %d  %d  %d\n ", 
                  f->vertex[0]->vnum, 
                  f->vertex[1]->vnum, 
                  f->vertex[2]->vnum);

     //     printf("%lf %lf %lf 1\n", (double)random()/MAX_INT,(double)random()/MAX_INT,(double)random()/MAX_INT);
     f = f->next;
   } while ( f != faces );  

   f=faces;


}

/*---------------------------------------------------------------------
Print: Prints out the vertices and the faces.  Uses the vnum indices 
corresponding to the order in which the vertices were input.
Output is in PostScript format.
---------------------------------------------------------------------*/
void	Print( void )
{
   /* Pointers to vertices, edges, faces. */
   tVertex  v;
   tEdge    e;
   tFace    f;
   int xmin, ymin, xmax, ymax;
   int a[3], b[3];  /* used to compute normal vector */
   /* Counters for Euler's formula. */
   int 	V = 0, E = 0 , F = 0;
   /* Note: lowercase==pointer, uppercase==counter. */

   /*-- find X min & max --*/
   v = vertices;
   xmin = xmax = v->v[X];
   do {
      if( v->v[X] > xmax ) xmax = v->v[X];
      else
	 if( v->v[X] < xmin ) xmin = v->v[X];
      v = v->next;
   } while ( v != vertices );
	
   /*-- find Y min & max --*/
   v = vertices;
   ymin = ymax = v->v[Y];
   do {
      if( v->v[Y] > ymax ) ymax = v->v[Y];
      else
	 if( v->v[Y] < ymin ) ymin = v->v[Y];
      v = v->next;
   } while ( v != vertices );
	
   /* PostScript header */
   printf("%%!PS\n");
   printf("%%%%BoundingBox: %d %d %d %d\n", 
	  xmin, ymin, xmax, ymax);
   printf(".00 .00 setlinewidth\n");
   printf("%d %d translate\n", -xmin+72, -ymin+72 );
   /* The +72 shifts the figure one inch from the lower left corner */

   /* Vertices. */
   v = vertices;
   do {                                 
      if( v->mark ) V++;           
      v = v->next;
   } while ( v != vertices );
   printf("\n%%%% Vertices:\tV = %d\n", V);
   printf("%%%% index:\tx\ty\tz\n");
   do {                                 
      printf( "%%%% %5d:\t%d\t%d\t%d\n", 
	     v->vnum, v->v[X], v->v[Y], v->v[Z] );
      v = v->next;
   } while ( v != vertices );
	
   /* Faces. */
   /* visible faces are printed as PS output */
   f = faces;
   do {
      ++F;                              
      f  = f ->next;
   } while ( f  != faces );
   printf("\n%%%% Faces:\tF = %d\n", F );
   printf("%%%% Visible faces only: \n");
   do {           
      /* Print face only if it is visible: if normal vector >= 0 */
      SubVec( f->vertex[1]->v, f->vertex[0]->v, a );
      SubVec( f->vertex[2]->v, f->vertex[1]->v, b );	  
      if(( a[0] * b[1] - a[1] * b[0] ) >= 0 )
      {
	 printf("%%%% vnums:  %d  %d  %d\n", 
		f->vertex[0]->vnum, 
		f->vertex[1]->vnum, 
		f->vertex[2]->vnum);
	 printf("newpath\n");
	 printf("%d\t%d\tmoveto\n", 
		f->vertex[0]->v[X], f->vertex[0]->v[Y] );
	 printf("%d\t%d\tlineto\n", 
		f->vertex[1]->v[X], f->vertex[1]->v[Y] );
	 printf("%d\t%d\tlineto\n", 
		f->vertex[2]->v[X], f->vertex[2]->v[Y] );
	 printf("closepath stroke\n\n");
      }
      f = f->next;
   } while ( f != faces );

   /* prints a list of all faces */
   printf("%%%% List of all faces: \n");
   printf("%%%%\tv0\tv1\tv2\t(vertex indices)\n");
   do {
      printf("%%%%\t%d\t%d\t%d\n",
	     f->vertex[0]->vnum,
	     f->vertex[1]->vnum,
	     f->vertex[2]->vnum );
      f = f->next;
   } while ( f != faces );
	
   /* Edges. */	
   e = edges;
   do {
      E++;
      e = e->next;
   } while ( e != edges );
   printf("\n%%%% Edges:\tE = %d\n", E );
   /* Edges not printed out (but easily added). */

   printf("\nshowpage\n\n");

   check = TRUE;
   CheckEuler( V, E, F );

}

/*---------------------------------------------------------------------
SubVec:  Computes a - b and puts it into c.
---------------------------------------------------------------------*/
void    SubVec( int a[3], int b[3], int c[3])
{
   int  i;

   for( i=0; i < 2; i++ )
      c[i] = a[i] - b[i];

}

/*---------------------------------------------------------------------
 DoubleTriangle builds the initial double triangle.  It first finds 3 
 noncollinear points and makes two faces out of them, in opposite order.
 It then finds a fourth point that is not coplanar with that face.  The  
 vertices are stored in the face structure in counterclockwise order so 
 that the volume between the face and the point is negative. Lastly, the
 3 newfaces to the fourth point are constructed and the data structures
 are cleaned up. 
---------------------------------------------------------------------*/
void    DoubleTriangle( void )
{
   tVertex  v0, v1, v2, v3, t;
   tFace    f0, f1 = NULL;
   tEdge    e0, e1, e2, s;
   int      vol;
	
   /* Find 3 noncollinear points. */
   v0 = vertices;
   while ( Collinear( v0, v0->next, v0->next->next ) )
      if ( ( v0 = v0->next ) == vertices )
         printf("DoubleTriangle:  All points are Collinear!\n"), exit(0);
   v1 = v0->next;
   v2 = v1->next;
	
   /* Mark the vertices as processed. */
   v0->mark = PROCESSED;
   v1->mark = PROCESSED;
   v2->mark = PROCESSED;
   
   /* Create the two "twin" faces. */
   f0 = MakeFace( v0, v1, v2, f1 );
   f1 = MakeFace( v2, v1, v0, f0 );

   /* Link adjacent face fields. */
   f0->edge[0]->adjface[1] = f1;
   f0->edge[1]->adjface[1] = f1;
   f0->edge[2]->adjface[1] = f1;
   f1->edge[0]->adjface[1] = f0;
   f1->edge[1]->adjface[1] = f0;
   f1->edge[2]->adjface[1] = f0;
	
   /* Find a fourth, noncoplanar point to form tetrahedron. */
   v3 = v2->next;
   vol = VolumeSign( f0, v3 );
   while ( !vol )   {
      if ( ( v3 = v3->next ) == v0 ) 
         printf("DoubleTriangle:  All points are coplanar!\n"), exit(0);
      vol = VolumeSign( f0, v3 );
   }
	
   /* Insure that v3 will be the first added. */
   vertices = v3;
   if ( debug ) {
      fprintf(stderr, "DoubleTriangle: finished. Head repositioned at v3.\n");
      PrintOut( vertices );
   }

	
}

  
/*---------------------------------------------------------------------
ConstructHull adds the vertices to the hull one at a time.  The hull
vertices are those in the list marked as onhull.
---------------------------------------------------------------------*/
void	ConstructHull( void )
{
   tVertex  v, vnext;
   int 	    vol;
   bool	    changed;	/* T if addition changes hull; not used. */

   v = vertices;
   do {
      vnext = v->next;
      if ( !v->mark ) {
         v->mark = PROCESSED;
	 changed = AddOne( v );
	 CleanUp();

	 if ( check ) {
	    fprintf(stderr,"ConstructHull: After Add of %d & Cleanup:\n", 
               v->vnum);
	    Checks();
	 }
	 if ( debug )
            PrintOut( v );
      }
      v = vnext;
   } while ( v != vertices );
}

/*---------------------------------------------------------------------
AddOne is passed a vertex.  It first determines all faces visible from 
that point.  If none are visible then the point is marked as not 
onhull.  Next is a loop over edges.  If both faces adjacent to an edge
are visible, then the edge is marked for deletion.  If just one of the
adjacent faces is visible then a new face is constructed.
---------------------------------------------------------------------*/
bool 	AddOne( tVertex p )
{
   tFace  f; 
   tEdge  e, temp;
   int 	  vol;
   bool	  vis = FALSE;

   if ( debug ) {
      fprintf(stderr, "AddOne: starting to add v%d.\n", p->vnum);
      PrintOut( vertices );
   }

   /* Mark faces visible from p. */
   f = faces;
   do {
      vol = VolumeSign( f, p );
      if (debug) fprintf(stderr, 
         "faddr: %6x   paddr: %6x   Vol = %d\n", f,p,vol);
      if ( vol < 0 ) {
	 f->visible = VISIBLE;  
	 vis = TRUE;                      
      }
      f = f->next;
   } while ( f != faces );

   /* If no faces are visible from p, then p is inside the hull. */
   if ( !vis ) {
      p->onhull = !ONHULL;  
      return FALSE; 
   }

   /* Mark edges in interior of visible region for deletion.
      Erect a newface based on each border edge. */
   e = edges;
   do {
      temp = e->next;
      if ( e->adjface[0]->visible && e->adjface[1]->visible )
	 /* e interior: mark for deletion. */
	 e->delete = REMOVED;
      else if ( e->adjface[0]->visible || e->adjface[1]->visible ) 
	 /* e border: make a new face. */
	 e->newface = MakeConeFace( e, p );
      e = temp;
   } while ( e != edges );
   return TRUE;
}

/*---------------------------------------------------------------------
VolumeSign returns the sign of the volume of the tetrahedron determined by f
and p.  VolumeSign is +1 iff p is on the negative side of f,
where the positive side is determined by the rh-rule.  So the volume 
is positive if the ccw normal to f points outside the tetrahedron.
The final fewer-multiplications form is due to Bob Williamson.
---------------------------------------------------------------------*/
int  VolumeSign( tFace f, tVertex p )
{
   double  vol;
   int     voli;
   double  ax, ay, az, bx, by, bz, cx, cy, cz;

   ax = f->vertex[0]->v[X] - p->v[X];
   ay = f->vertex[0]->v[Y] - p->v[Y];
   az = f->vertex[0]->v[Z] - p->v[Z];
   bx = f->vertex[1]->v[X] - p->v[X];
   by = f->vertex[1]->v[Y] - p->v[Y];
   bz = f->vertex[1]->v[Z] - p->v[Z];
   cx = f->vertex[2]->v[X] - p->v[X];
   cy = f->vertex[2]->v[Y] - p->v[Y];
   cz = f->vertex[2]->v[Z] - p->v[Z];

   vol =   ax * (by*cz - bz*cy)
         + ay * (bz*cx - bx*cz)
         + az * (bx*cy - by*cx);

   if ( debug )
      fprintf(stderr,"Face=%6x; Vertex=%d: vol(int) = %d, vol(double) = %lf\n",
	      f,p->vnum,voli,vol);

   /* The volume should be an integer. */
   if      ( vol >  0.5 )  return  1;
   else if ( vol < -0.5 )  return -1;
   else                    return  0;
}
/*---------------------------------------------------------------------*/
int  Volumei( tFace f, tVertex p )
{
   double  vol;
   int     voli;
   double  ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz;

   ax = f->vertex[0]->v[X] - p->v[X];
   ay = f->vertex[0]->v[Y] - p->v[Y];
   az = f->vertex[0]->v[Z] - p->v[Z];
   bx = f->vertex[1]->v[X] - p->v[X];
   by = f->vertex[1]->v[Y] - p->v[Y];
   bz = f->vertex[1]->v[Z] - p->v[Z];
   cx = f->vertex[2]->v[X] - p->v[X];
   cy = f->vertex[2]->v[Y] - p->v[Y];
   cz = f->vertex[2]->v[Z] - p->v[Z];

   vol =  (ax * (by*cz - bz*cy)
         + ay * (bz*cx - bx*cz)
         + az * (bx*cy - by*cx));

   if ( debug )
      fprintf(stderr,"Face=%6x; Vertex=%d: vol(int) = %d, vol(double) = %lf\n",
	      f,p->vnum,voli,vol);

   /* The volume should be an integer. */
   if      ( vol > 0.5 )   return  1;
   else if ( vol < -0.5 )  return -1;
   else                    return  0;
}


/*-------------------------------------------------------------------*/
void	PrintPoint( tVertex p )
{
   int	i;

   for ( i = 0; i < 3; i++ )
      printf("\t%d", p->v[i]);
   putchar('\n');
}


/*---------------------------------------------------------------------
Computes the z-coordinate of the vector normal to face f.
---------------------------------------------------------------------*/
int	Normz( tFace f )
{
   tVertex a, b, c;

   a = f->vertex[0];
   b = f->vertex[1];
   c = f->vertex[2];

   return 
      ( b->v[X] - a->v[X] ) * ( c->v[Y] - a->v[Y] ) -
      ( b->v[Y] - a->v[Y] ) * ( c->v[X] - a->v[X] );
}

/*---------------------------------------------------------------------
MakeConeFace makes a new face and two new edges between the 
edge and the point that are passed to it. It returns a pointer to
the new face.
---------------------------------------------------------------------*/
tFace	MakeConeFace( tEdge e, tVertex p )
{
   tEdge  new_edge[2];
   tFace  new_face;
   int 	  i, j;

   /* Make two new edges (if don't already exist). */
   for ( i=0; i < 2; ++i ) 
      /* If the edge exists, copy it into new_edge. */
      if ( !( new_edge[i] = e->endpts[i]->duplicate) ) {
	 /* Otherwise (duplicate is NULL), MakeNullEdge. */
	 new_edge[i] = MakeNullEdge();
	 new_edge[i]->endpts[0] = e->endpts[i];
	 new_edge[i]->endpts[1] = p;
	 e->endpts[i]->duplicate = new_edge[i];
      }

   /* Make the new face. */
   new_face = MakeNullFace();   
   new_face->edge[0] = e;
   new_face->edge[1] = new_edge[0];
   new_face->edge[2] = new_edge[1];
   MakeCcw( new_face, e, p ); 
        
   /* Set the adjacent face pointers. */
   for ( i=0; i < 2; ++i )
      for ( j=0; j < 2; ++j )  
	 /* Only one NULL link should be set to new_face. */
	 if ( !new_edge[i]->adjface[j] ) {
	    new_edge[i]->adjface[j] = new_face;
	    break;
	 }
        
   return new_face;
}

/*---------------------------------------------------------------------
MakeCcw puts the vertices in the face structure in counterclock wise 
order.  We want to store the vertices in the same 
order as in the visible face.  The third vertex is always p.
---------------------------------------------------------------------*/
void	MakeCcw( tFace f, tEdge e, tVertex p )
{
   tFace  fv;   /* The visible face adjacent to e */
   int    i;    /* Index of e->endpoint[0] in fv. */
   tEdge  s;	/* Temporary, for swapping */
      
   if  ( e->adjface[0]->visible )      
        fv = e->adjface[0];
   else fv = e->adjface[1];
       
   /* Set vertex[0] & [1] of f to have the same orientation
      as do the corresponding vertices of fv. */ 
   for ( i=0; fv->vertex[i] != e->endpts[0]; ++i )
      ;
   /* Orient f the same as fv. */
   if ( fv->vertex[ (i+1) % 3 ] != e->endpts[1] ) {
      f->vertex[0] = e->endpts[1];  
      f->vertex[1] = e->endpts[0];    
   }
   else {                               
      f->vertex[0] = e->endpts[0];   
      f->vertex[1] = e->endpts[1];      
      SWAP( s, f->edge[1], f->edge[2] );
   }
   /* This swap is tricky. e is edge[0]. edge[1] is based on endpt[0],
      edge[2] on endpt[1].  So if e is oriented "forwards," we
      need to move edge[1] to follow [0], because it precedes. */
   
   f->vertex[2] = p;
}
 
/*---------------------------------------------------------------------
MakeNullEdge creates a new cell and initializes all pointers to NULL
and sets all flags to off.  It returns a pointer to the empty cell.
---------------------------------------------------------------------*/
tEdge 	MakeNullEdge( void )
{
   tEdge  e;

   NEW( e, tsEdge );
   e->adjface[0] = e->adjface[1] = e->newface = NULL;
   e->endpts[0] = e->endpts[1] = NULL;
   e->delete = !REMOVED;
   ADD( edges, e );
   return e;
}

/*--------------------------------------------------------------------
MakeNullFace creates a new face structure and initializes all of its
flags to NULL and sets all the flags to off.  It returns a pointer
to the empty cell.
---------------------------------------------------------------------*/
tFace 	MakeNullFace( void )
{
   tFace  f;
   int    i;

   NEW( f, tsFace);
   for ( i=0; i < 3; ++i ) {
      f->edge[i] = NULL;
      f->vertex[i] = NULL;
   }
   f->visible = !VISIBLE;
   ADD( faces, f );
   return f;
}

/*---------------------------------------------------------------------
MakeFace creates a new face structure from three vertices (in ccw
order).  It returns a pointer to the face.
---------------------------------------------------------------------*/
tFace   MakeFace( tVertex v0, tVertex v1, tVertex v2, tFace fold )
{
   tFace  f;
   tEdge  e0, e1, e2;

   /* Create edges of the initial triangle. */
   if( !fold ) {
     e0 = MakeNullEdge();
     e1 = MakeNullEdge();
     e2 = MakeNullEdge();
   }
   else { /* Copy from fold, in reverse order. */
     e0 = fold->edge[2];
     e1 = fold->edge[1];
     e2 = fold->edge[0];
   }
   e0->endpts[0] = v0;              e0->endpts[1] = v1;
   e1->endpts[0] = v1;              e1->endpts[1] = v2;
   e2->endpts[0] = v2;              e2->endpts[1] = v0;
	
   /* Create face for triangle. */
   f = MakeNullFace();
   f->edge[0]   = e0;  f->edge[1]   = e1; f->edge[2]   = e2;
   f->vertex[0] = v0;  f->vertex[1] = v1; f->vertex[2] = v2;
	
   /* Link edges to face. */
   e0->adjface[0] = e1->adjface[0] = e2->adjface[0] = f;
	
   return f;
}

/*---------------------------------------------------------------------
CleanUp goes through each data structure list and clears all
flags and NULLs out some pointers.  The order of processing
(edges, faces, vertices) is important.
---------------------------------------------------------------------*/
void	CleanUp( void )
{
   CleanEdges();
   CleanFaces();
   CleanVertices();
}

/*---------------------------------------------------------------------
CleanEdges runs through the edge list and cleans up the structure.
If there is a newface then it will put that face in place of the 
visible face and NULL out newface. It also deletes so marked edges.
---------------------------------------------------------------------*/
void	CleanEdges( void )
{
   tEdge  e;	/* Primary index into edge list. */
   tEdge  t;	/* Temporary edge pointer. */
		
   /* Integrate the newface's into the data structure. */
   /* Check every edge. */
   e = edges;
   do {
      if ( e->newface ) { 
	 if ( e->adjface[0]->visible )
	    e->adjface[0] = e->newface; 
	 else	e->adjface[1] = e->newface;
	 e->newface = NULL;
      }
      e = e->next;
   } while ( e != edges );

   /* Delete any edges marked for deletion. */
   while ( edges && edges->delete ) { 
      e = edges;
      DELETE( edges, e );
   }
   e = edges->next;
   do {
      if ( e->delete ) {
	 t = e;
	 e = e->next;
	 DELETE( edges, t );
      }
      else e = e->next;
   } while ( e != edges );
}

/*---------------------------------------------------------------------
CleanFaces runs through the face list and deletes any face marked visible.
---------------------------------------------------------------------*/
void	CleanFaces( void )
{
   tFace  f;	/* Primary pointer into face list. */
   tFace  t;	/* Temporary pointer, for deleting. */
	

   while ( faces && faces->visible ) { 
      f = faces;
      DELETE( faces, f );
   }
   f = faces->next;
   do {
      if ( f->visible ) {
	 t = f;
	 f = f->next;
	 DELETE( faces, t );
      }
      else f = f->next;
   } while ( f != faces );
}

/*---------------------------------------------------------------------
CleanVertices runs through the vertex list and deletes the 
vertices that are marked as processed but are not incident to any 
undeleted edges. 
---------------------------------------------------------------------*/
void	CleanVertices( void )
{
   tEdge    e;
   tVertex  v, t;
	
   /* Mark all vertices incident to some undeleted edge as on the hull. */
   e = edges;
   do {
      e->endpts[0]->onhull = e->endpts[1]->onhull = ONHULL;
      e = e->next;
   } while (e != edges);
	
   /* Delete all vertices that have been processed but
      are not on the hull. */
   while ( vertices && vertices->mark && !vertices->onhull ) { 
      v = vertices;
      DELETE( vertices, v );
   }
   v = vertices->next;
   do {
      if ( v->mark && !v->onhull ) {    
	 t = v; 
	 v = v->next;
	 DELETE( vertices, t )
      }
      else v = v->next;
   } while ( v != vertices );
	
   /* Reset flags. */
   v = vertices;
   do {
      v->duplicate = NULL; 
      v->onhull = !ONHULL; 
      v = v->next;
   } while ( v != vertices );
}

/*---------------------------------------------------------------------
Collinear checks to see if the three points given are collinear,
by checking to see if each element of the cross product is zero.
---------------------------------------------------------------------*/
bool	Collinear( tVertex a, tVertex b, tVertex c )
{
   return 
         ( c->v[Z] - a->v[Z] ) * ( b->v[Y] - a->v[Y] ) -
         ( b->v[Z] - a->v[Z] ) * ( c->v[Y] - a->v[Y] ) == 0
      && ( b->v[Z] - a->v[Z] ) * ( c->v[X] - a->v[X] ) -
         ( b->v[X] - a->v[X] ) * ( c->v[Z] - a->v[Z] ) == 0
      && ( b->v[X] - a->v[X] ) * ( c->v[Y] - a->v[Y] ) -
         ( b->v[Y] - a->v[Y] ) * ( c->v[X] - a->v[X] ) == 0  ;
}

/*---------------------------------------------------------------------
Consistency runs through the edge list and checks that all
adjacent faces have their endpoints in opposite order.  This verifies
that the vertices are in counterclockwise order.
---------------------------------------------------------------------*/
void	Consistency( void )
{
   register tEdge  e;
   register int    i, j;

   e = edges;

   do {
      /* find index of endpoint[0] in adjacent face[0] */
      for ( i = 0; e->adjface[0]->vertex[i] != e->endpts[0]; ++i )
	 ;
   
      /* find index of endpoint[0] in adjacent face[1] */
      for ( j = 0; e->adjface[1]->vertex[j] != e->endpts[0]; ++j )
	 ;

      /* check if the endpoints occur in opposite order */
      if ( !( e->adjface[0]->vertex[ (i+1) % 3 ] ==
	      e->adjface[1]->vertex[ (j+2) % 3 ] ||
	      e->adjface[0]->vertex[ (i+2) % 3 ] ==
	      e->adjface[1]->vertex[ (j+1) % 3 ] )  )
	 break;
      e = e->next;

   } while ( e != edges );

   if ( e != edges )
      fprintf( stderr, "Checks: edges are NOT consistent.\n");
   else
      fprintf( stderr, "Checks: edges consistent.\n");

}

/*---------------------------------------------------------------------
Convexity checks that the volume between every face and every
point is negative.  This shows that each point is inside every face
and therefore the hull is convex.
---------------------------------------------------------------------*/
void	Convexity( void )
{
   register tFace    f;
   register tVertex  v;
   int               vol;

   f = faces;
   
   do {
      v = vertices;
      do {
	 if ( v->mark ) {
	    vol = VolumeSign( f, v );
	    if ( vol < 0 )
	       break;
	 }
	 v = v->next;
      } while ( v != vertices );

      f = f->next;

   } while ( f != faces );

   if ( f != faces )
      fprintf( stderr, "Checks: NOT convex.\n");
   else if ( check ) 
      fprintf( stderr, "Checks: convex.\n");
}

/*---------------------------------------------------------------------
CheckEuler checks Euler's relation, as well as its implications when
all faces are known to be triangles.  Only prints positive information
when debug is true, but always prints negative information.
---------------------------------------------------------------------*/
void	CheckEuler( int V, int E, int F )
{
   if ( check )
      fprintf( stderr, "Checks: V, E, F = %d %d %d:\t", V, E, F);

   if ( (V - E + F) != 2 )
      fprintf( stderr, "Checks: V-E+F != 2\n");
   else if ( check )
      fprintf( stderr, "V-E+F = 2\t");


   if ( F != (2 * V - 4) )
      fprintf( stderr, "Checks: F=%d != 2V-4=%d; V=%d\n",
	      F, 2*V-4, V);
   else if ( check ) 
      fprintf( stderr, "F = 2V-4\t");
   
   if ( (2 * E) != (3 * F) )
      fprintf( stderr, "Checks: 2E=%d != 3F=%d; E=%d, F=%d\n",
	      2*E, 3*F, E, F );
   else if ( check ) 
      fprintf( stderr, "2E = 3F\n");
}

/*-------------------------------------------------------------------*/
void	Checks( void )
{
   tVertex  v;
   tEdge    e;
   tFace    f;
   int 	   V = 0, E = 0 , F = 0;

   Consistency();
   Convexity();
   if ( v = vertices )
      do {
         if (v->mark) V++;
	 v = v->next;
      } while ( v != vertices );
   if ( e = edges )
      do {
         E++;
	 e = e->next;
      } while ( e != edges );
   if ( f = faces )
      do {
         F++;
	 f  = f ->next;
      } while ( f  != faces );
   CheckEuler( V, E, F );
}


/*===================================================================
These functions are used whenever the debug flag is set.
They print out the entire contents of each data structure.  
Printing is to standard error.  To grab the output in a file in the csh, 
use this:
	chullsp < i.file >&! o.file
=====================================================================*/
/*-------------------------------------------------------------------*/
void	PrintOut( tVertex v )
{
   fprintf( stderr, "\nHead vertex %d = %6x :\n", v->vnum, v );
   PrintVertices();
   PrintEdges();
   PrintFaces();
}

/*-------------------------------------------------------------------*/
void	PrintVertices( void )
{
   tVertex  temp;

   temp = vertices;
   fprintf (stderr, "Vertex List\n");
   if (vertices) do {
      fprintf(stderr,"  addr %6x\t", vertices );
      fprintf(stderr,"  vnum %4d", vertices->vnum );
      fprintf(stderr,"   (%6d,%6d,%6d)",vertices->v[X],
	      vertices->v[Y], vertices->v[Z] );
      fprintf(stderr,"   active:%3d", vertices->onhull );
      fprintf(stderr,"   dup:%5x", vertices->duplicate );
      fprintf(stderr,"   mark:%2d\n", vertices->mark );
      vertices = vertices->next;
   } while ( vertices != temp );

}

/*-------------------------------------------------------------------*/
void	PrintEdges( void )
{
   tEdge  temp;
   int 	  i;
	
   temp = edges;
   fprintf (stderr, "Edge List\n");
   if (edges) do {
      fprintf( stderr, "  addr: %6x\t", edges );
      fprintf( stderr, "adj: ");
      for (i=0; i<2; ++i) 
	 fprintf( stderr, "%6x", edges->adjface[i] );
      fprintf( stderr, "  endpts:");
      for (i=0; i<2; ++i) 
	 fprintf( stderr, "%4d", edges->endpts[i]->vnum);
      fprintf( stderr, "  del:%3d\n", edges->delete );
      edges = edges->next; 
   } while (edges != temp );

}

/*-------------------------------------------------------------------*/
void	PrintFaces( void )
{
   int 	  i;
   tFace  temp;

   temp = faces;
   fprintf (stderr, "Face List\n");
   if (faces) do {
      fprintf(stderr, "  addr: %6x\t", faces );
      fprintf(stderr, "  edges:");
      for( i=0; i<3; ++i )
	 fprintf(stderr, "%6x", faces->edge[i] );
      fprintf(stderr, "  vert:");
      for ( i=0; i<3; ++i)
	 fprintf(stderr, "%4d", faces->vertex[i]->vnum );
      fprintf(stderr, "  vis: %d\n", faces->visible );
      faces= faces->next;
   } while ( faces != temp );

}


